
#    Note:  This is loaded into the namespace  MEISAPSR

# variable started 0

#Analyser.Current+             : -a=0x00009000 -c3 -g=1 -q=u -u=A -d=0 -r=0x0000:0uA,0x0fff:1000uA
#Analyser.Current-             : -a=0x00009000 -c4 -g=1 -q=u -u=A -d=0 -r=0x0000:0uA,0x0fff:1000uA
#Analyser.Energy               : -a0x00008000 -o0x00000088 -w16 -d16 -ud -mn -prw -qk -teV -n3 \
#                                    -r0x0:0.000keV,0xffff:416.667keV -i0.000keV
#Analyser.Power                : -a0x00008000 -o0x0000008a -w16 -d8 -ud -mn -prw -l1 -b0 -i0 -zd -s0
#Analyser.Volts+               : -a=0x00009000 -c1 -g=1 -q=k -u=V -d=2 -r=0x0000:0.00kV,0x0fff:25.00kV
#Analyser.Volts-               : -a=0x00009000 -c2 -g=1 -q=k -u=V -d=2 -r=0x0000:0.00kV,0x0fff:25.00kV

#xvme560   0x9000
#xvme240   0x8000

if {![info exists started]} {variable started 0}

if {$started != 0} {

     variable JS ""
     variable LogFlag

     if {$LogFlag == 1} {InsertLog "Last Updated: [clock format [clock seconds] -format "%B %d, %Y %H:%M:%S"]"}

    do_click
}

if {$started == 0} {           ;#  first time load code and define local variables
    set started 1

    set Color1(0) Red
    set Color1(1) Green
    set Color2(0) Cyan
    set Color2(1) #D4A017     ;# Orange Gold

    variable ShowLogFlag 0

    variable LogText ""
    variable LogFlag 0
    variable LogCount 0

    variable JS ""
    variable HelpTopic "undefined"

    variable OFF 1
    variable ON 0

    variable STOP 1
    variable GO 0

    variable RAMP 10
    variable rampafterID ""

    variable Offset0 0  ;# threshold
    variable Offset1 0.0
    variable Offset2 0.0
    variable Offset3 0.0

    variable UNIT0 1
    variable UNIT1 0
    variable UNIT2 0
    variable UNIT3 0

    variable multiplier 60
    variable threshold 50000
    variable ramptime 10
    variable rampstep 1
    variable going 0
    variable steptime 10    ;#  move at 10 seconds/step
    variable final 0
    variable latest 0
    variable current 0


   proc  show_variables {} {
   
       variable OFF
       variable ON
       variable Offset0
       variable Offset1
       variable Offset2
       variable Offset3
    variable multiplier
    variable threshold
    variable ramptime
    variable final
    variable current
    variable going
     
   set registerList [list Analyser.Energy Analyser.Volts+]

       InsertLog "$OFF $ON $Offset0 $Offset1 $Offset2 $Offset3"
       InsertLog "final=$final; threshold=$threshold; current=$current; ramptime=$ramptime; multiplier=$multiplier; going=$going"

      foreach register $registerList {
         InsertLog "$register => [ReadMEISRegister $register]"
      }
      
   }

   proc Help {subject} {
       variable HelpTopic

       switch $subject {

          Introduction -
          "Introduction to the System"  {set HelpTopic UsingMEISDataAcq}

          default {}
       }

       do_help
   }
   
   proc do_ramping {vl} {
    variable rampstep
    variable going
    variable steptime    ;#  move at 10 seconds/step
    variable threshold
    variable final
    variable latest
    variable current
    variable multiplier
    variable rampafterID
   
      
   if {$vl == 1} then {    ;# go
      do_redisplay
 puts "stopgo current: $current eV"
 puts "stopgo threshold: $threshold eV"
 puts "stopgo final: $final eV"
      if {$final < $threshold || $final < $current} then {
         do_Energy $final
         set going 0
         do_continue_ramping
         return
      }

      if {$current < $threshold} then {
         do_Energy $threshold
 puts "current now: $current eV"
         set pause $steptime
      } else {
         set pause 0
      }

      set latest $current
      set period [expr $ramptime * $multiplier]

 puts "total period: $period seconds"
      set steps [expr $period / $steptime]

      if {$steps < 1} then {set steps 1}
 puts "No. steps: $steps"
      set interval [expr $final - $latest]
 puts "interval: $interval eV"
      set rampstep [expr $interval / $steps]
      if {$rampstep < 1} then {set rampstep 1}
 puts "ramp per step: $rampstep"

      set going 1
 puts "setting afterID with pause $pause"
      set rampafterID [after [expr $pause * 1000] ::MEISAPSR::do_continue_ramping]
      SetFooter "Analyser Energy Ramping Starting"
    } else {
       set going 0
       do_continue_ramping
    }
       
   }  
   
   proc do_continue_ramping {} {
 
    variable rampstep
    variable going
    variable steptime    ;#  move at 10 seconds/step
    variable final
    variable latest
    variable current
    variable rampafterID
    
    puts "do_continue_ramping"
    
    if {$rampafterID != ""} then {catch {kill-after $rampafterID}}

    if {$going == 1} then {
       set rampafterID [after [expr $steptime * 1000] ::MEISAPSR::do_continue_ramping]
  puts "Analyser Energy Ramping in Progress" 
       SetFooter "Analyser Energy Ramping in Progress" 
       do_redisplay
  puts "continue latest: $latest"
       set new [expr $latest + $rampstep]
       if {$new > $final} {set new $final}
  puts "continue new: $new"
  puts "continue using new: [expr round($new)]"
       do_Energy [expr round($new)]
       if {$new == $final} {set going 0}
       set latest $new
  puts "continue current now: $current"
  puts "continue latest now: $latest"
  puts "continue done"
     } else {
       do_redisplay
  puts "Analyser Energy Ramping Complete"
       SetFooter "Analyser Energy Ramping Complete"
     }
   }

   proc do_Setting {reg vl} {InsertLog "Setting $reg $vl"
      variable Offset0
      variable Offset1
      variable Offset2
      variable Offset3

# note  -  sliders do not return units for analogue registers

      variable threshold
      variable final

       switch $reg {
           rangeInput0   {set Offset0 $vl; variable threshold; set threshold [expr $Offset0 * 1000];do_Energy $Offset0}
           rangeInput1   {set Offset1 $vl; variable final; set final [expr $Offset1 * 1000]}
           default {InsertLog "AnalyserPSR: do_Setting: unexpected request $reg $vl"}
       }
   }
   
   proc do_Energy {vl} {WriteMEISRegister Analyser.Energy $vl}

   proc do_Power {vl} {    ;InsertLog "Power $vl"
      variable OFF
      variable ON

       switch $vl {
         0 {
            set OFF 1
            set ON 0
            WriteMEISRegister Analyser.Power 0
         }
         1 {
            set OFF 0
            set ON 1
            WriteVMERegister Analyser.Power 1
         }
         default {}
       }
   }

   proc do_Units {vl} {    ;InsertLog "Units $vl"
      variable UNIT0
      variable UNIT1
      variable UNIT2
      variable UNIT3

      variable multiplier

      set UNIT0 0; set UNIT1 0; set UNIT2 0; set UNIT3 0
      set UNIT$vl 1

      set multiplier [lindex [list 1 60 3600 86400] $vl]
   }

   proc do_stopgo {vl} {    ;InsertLog "stopgo $vl"
      variable STOP
      variable GO
      variable going

       switch $vl {
         0 {
            set STOP 1
            set GO 0
         }
         1 {
            set STOP 0
            set GO 1
         }
         default {}
       }
       
      if {$vl == $going} {return}
      
      set going $vl
      do_ramping $vl
      
   }
      

   proc do_Ramp {vl} {    ;InsertLog "Ramp $vl"
      variable RAMP
      variable ramptime

      set RAMP $vl
      set ramptime $vl

   }
   
   proc do_redisplay {} {
       variable going
       variable STOP
       variable GO

       
       UpdateModule
      
       switch $going {
         0 {
            set STOP 1
            set GO 0
         }
         1 {
            set STOP 0
            set GO 1
         }
         default {}
       }

   }


    proc UpdateModule {} {
      variable Offset0
      variable Offset1
      variable Offset2
      variable Offset3
      variable current

#    Offset0 => Threshold
#    Offset1 => End Energy
#    Offset2 => Set Energy
#    Offset3 => Voltage

#    Analyser.Energy    -r 0:0V,0xffff:416.667eV -i 0.0kV"

#    Analyser.Volts+    -r0:0V,0xfff:25kV"
#    Analyser.Volts-    -r0:0V,0xfff:25kV"
#    Analyser.Current+  -r0:0A,0xfff:1mA"
#    Analyser.Current-	-r0:0A,0xfff:1mA"

      set vl [stov [ReadMEISRegister Analyser.Energy]]
      set Offset2 [expr round($vl)]
      
      set vl [stov [ReadMEISRegister Analyser.Volts+]]
      set Offset3 [expr round($vl)]
      
      set current [expr $Offset2 * 1000]  ;# convert from KeV to eV

       variable ON
       variable OFF

         if {[expr [ReadMEISRegister Analyser.Power] & 1] == 0} {set ON 0; set OFF 1} else {set ON 1; set OFF 0}
   }

   proc  do_click {}  {
       variable cgidata
       variable started
       variable LogFlag
       variable ShowLogFlag

       set cgidata [ncgi::nvlist]

       if {$LogFlag == 1} {InsertLog "cgidata= ([llength $cgidata]) $cgidata"}

       if {[llength $cgidata] == 0} then {return}

       if {[llength $cgidata] < 2} then {
           InsertLog "CGI data is too short: $cgidata"
       } else {
           if {[lindex $cgidata 0] != "Widget"} then {
               InsertLog "Invalid CGI data: $cgidata"
           } else {           

              set w [first [lindex $cgidata 1]]
              set n [second [lindex $cgidata 1]]
              set vl [third [lindex $cgidata 1]]

               switch -glob $w {

                   RESET    {set started 0}
                   CLEAR    {EmptyLog}
                   ELOG     {PrintLog}
                   RELOAD   {}
                   VARIABLES {show_variables}
                   HELP     {Help $vl}

                   HIDE     {set ShowLogFlag 0}
                   SHOW     {set ShowLogFlag 1}
                   LOGON    {set LogFlag 1}
                   LOGOFF   {set LogFlag 0}

                   RANGE   -
                   TEXT    -
                   SETTING {do_Setting $n $vl}

                   POWER   {do_Power [second [lindex $cgidata 1]]}
                   UNIT    {do_Units [second [lindex $cgidata 1]]}
                   GO      {do_stopgo [second [lindex $cgidata 1]]}
                   RAMP    {do_Ramp [second [lindex $cgidata 1]]}

                   REDISPLAY {UpdateModule}
                   UPDATE    {UpdateModule}

                   EXPERT   {$vl [third [lindex $cgidata 1]]}

                   MainLog {}
                   undefined {}
                   default {InsertLog "unexpected CGI data:  $cgidata"}
               }
           }
        }
        UpdateModule
     }
     UpdateModule

}   ;# end of if {$started == 0}




