

#    Note:  This is loaded into the namespace  MEISDPSR

#variable started 0

#Detector.Current              : -a=0x00009000 -c6 -g=0 -q=u -u=A -d=0 -r=0x0000:0uA,0x0fff:1000uA
#Detector.Current.Request      : -a0x0000a000 -o0x00000002 -w16 -d16 -ud -mn -pwo -qu -tA -n0 -r0x0:0uA,0xfff:1000uA -i0uA
#Detector.Volts                : -a=0x00009000 -c5 -g=0 -q=  -u=V -d=0 -r=0x0000:0 V,0x0fff:3500 V
#Detector.Volts.Request        : -a0x0000a000 -o0x00000000 -w16 -d16 -ud -mn -pwo -q  -tV -n0 -r0x0:0 V,0xfff:3500 V -i0 V

#xvme560   0x9000
#msda12    0xa000


if {![info exists started]} {variable started 0}

if {$started != 0} {

     variable JS ""
     variable LogFlag

     if {$LogFlag == 1} {InsertLog "Last Updated: [clock format [clock seconds] -format "%B %d, %Y %H:%M:%S"]"}

    do_click
}

if {$started == 0} {           ;#  first time load code and define local variables
    set started 1

    set Color1(0) Red
    set Color1(1) Green
    set Color2(0) Cyan
    set Color2(1) #D4A017     ;# Orange Gold

    variable ShowLogFlag 0

    variable LogText ""
    variable LogFlag 0
    variable LogCount 0

    variable JS ""
    variable HelpTopic "undefined"

    variable OFF 1
    variable ON 0

    variable STOP 1
    variable GO 0

    variable RAMP 10
    
    variable rampafterID ""

    variable Offset0 0  ;# threshold
    variable Offset1 0
    variable Offset2 0
    variable Offset3 0

    variable UNIT0 1
    variable UNIT1 0
    variable UNIT2 0
    variable UNIT3 0

    variable multiplier 60
    variable threshold 1200
    variable ramptime 10
    variable rampstep 1
    variable going 0
    variable steptime 10    ;#  move at 10 seconds/step
    variable final 0
    variable latest 0
    variable current 0


   proc  show_variables {} {
      variable Offset0
      variable Offset1
      variable Offset2
      variable Offset3

      variable OFF
      variable ON
      
    variable multiplier
    variable threshold
    variable ramptime
    variable final
    variable current
    variable going

   set registerList [list  Detector.Volts Detector.Current]
   
       InsertLog "$OFF $ON $Offset0 $Offset1 $Offset2 $Offset3"
       InsertLog "final=$final; threshold=$threshold; current=$current; ramptime=$ramptime; multiplier=$multiplier; going=$going"

      foreach register $registerList {
         InsertLog "$register => [ReadMEISRegister $register]"
      }

   }

   proc Help {subject} {
       variable HelpTopic

       switch $subject {

          Introduction -
          "Introduction to the System"  {set HelpTopic UsingMEISDataAcq}
          default {}
       }

       do_help
   }
   
   proc do_ramping {vl} {

    variable rampstep
    variable steptime    ;#  move at 10 seconds/step
    variable ramptime
    variable threshold
    variable final
    variable latest
    variable current
    variable multiplier
    variable going
    variable rampafterID
    
    
    if {$vl == 1} then {   ;#  go
      do_redisplay
 puts "stopgo current: $current V"
 puts "stopgo threshold: $threshold V"
 puts "stopgo final: $final V"
      if {$final < $threshold || $final < $current} then {
         do_Volts $final
         set going 0
         do_continue_ramping
         return
      }

      if {$current < $threshold} then {
         do_Volts $threshold
 puts "current now: $current V"
         set pause $steptime
      } else {
         set pause 0
      }

      set latest $current
      set period [expr $ramptime * $multiplier]

 puts "total period: $period seconds"
      set steps [expr $period / $steptime]

      if {$steps < 1} then {set steps 1}
 puts "No. steps: $steps"
      set interval [expr $final - $latest]
 puts "interval: $interval V"
      set rampstep [expr $interval / $steps]
      if {$rampstep < 1} then {set rampstep 1}
 puts "ramp per step: $rampstep"

      set going 1
 puts "setting afterID with pause $pause"
      set rampafterID [after [expr $pause * 1000] ::MEISDPSR::do_continue_ramping]
      SetFooter "Detector Voltage Ramping Starting"
    } else {      ;#  stop
       set going 0
       do_continue_ramping
    }    
   
   }
   
   proc do_continue_ramping {} {
 
    variable rampstep
    variable going
    variable steptime    ;#  move at 10 seconds/step
    variable final
    variable latest
    variable current
    variable rampafterID
    
    puts "do_continue_ramping"
    
    if {$rampafterID != ""} then {catch {kill-after $rampafterID}}

    if {$going == 1} then {
       set rampafterID [after [expr $steptime * 1000] ::MEISDPSR::do_continue_ramping]
  puts "Detector Voltage Ramping in Progress" 
       SetFooter "Detector Voltage Ramping in Progress" 
       do_redisplay
  puts "continue latest: $latest"
       set new [expr $latest + $rampstep]
       if {$new > $final} {set new $final}
  puts "continue new: $new"
  puts "continue using new: [expr round($new)]"
       do_Volts [expr round($new)]
       if {$new == $final} {set going 0}
       set latest $new
  puts "continue current now: $current"
  puts "continue latest now: $latest"
  puts "continue done"
     } else {
       do_redisplay
  puts "Detector Voltage Ramping Complete"
       SetFooter "Detector Voltage Ramping Complete"
     }
   }

   proc do_Setting {reg vl} {        ;InsertLog "Setting $reg $vl"
      variable Offset0
      variable Offset1
      variable Offset2
      variable Offset3

#    Offset0 => threshold
#    Offset1 => final volts
#    Offset2 => Detector.Volts.Request
#    Offset3 => Detector.Volts

#Detector.Volts.Request      -r0:0V,0xfff:3.5kV  (3500V)   (MSDA12)
#Detector.Volts              -r0:0V,0xfff:3.5kV  (3500V)   (xvme560)

      variable threshold
      variable final

       switch $reg {
           rangeInput0   {set Offset0 $vl; variable threshold; set threshold $vl}
           rangeInput1   {set Offset1 $vl; variable final; set final $vl}
           rangeInput2   {set Offset2 $vl; do_Volts $vl}
           default       {InsertLog "do_Setting: unexpected request $reg $vl"}
       }
   }
   
  proc do_Volts {vl} {
      variable Offset2
      set Offset2 $vl
      WriteMEISRegister Detector.Volts.Request "[set vl]V"
  }

   proc do_Units {vl} {    ;InsertLog "Units $vl"
      variable UNIT0
      variable UNIT1
      variable UNIT2
      variable UNIT3

      variable multiplier

      set UNIT0 0; set UNIT1 0; set UNIT2 0; set UNIT3 0
      set UNIT$vl 1

      set multiplixer [lindex [list 1 60 3600 86400] $vl]
   }

   proc do_Ramp {vl} {
      variable RAMP
      variable ramptime

      set RAMP $vl
      set ramptime $vl

   }
   
   proc do_redisplay {} {
       variable STOP
       variable GO
       variable going
       
       UpdateModule
      
       switch $going {
         0 {
            set STOP 1
            set GO 0
         }
         1 {
            set STOP 0
            set GO 1
         }
         default {}
       }
       
       

   }

    proc UpdateModule {} {
       variable Offset0
       variable Offset1
       variable Offset2
       variable Offset3
       
       variable current

#    Offset0 => Threshold
#    Offset1 => End Energy   (final)
#    Offset2 => set Voltage  (Detector.Volts.Request)
#    Offset3 => monitor      (Detector.Volts)


 #     set Offset2 [ReadMEISRegister Detector.Volts.Request]
      set Offset3 [stov [ReadMEISRegister Detector.Volts]]
    
      set current $Offset2
   }

   proc do_stopgo {vl} {
      variable STOP
      variable GO
      variable going

       switch $vl {
         0 {
            set STOP 1
            set GO 0
         }
         1 {
            set STOP 0
            set GO 1
         }
         default {}
       }
       
      if {$vl == $going} {return}
      
      set going $vl
      do_ramping $vl
       
   }

   proc  do_click {}  {
       variable cgidata
       variable started
       variable LogFlag
       variable ShowLogFlag

       set cgidata [ncgi::nvlist]

       if {$LogFlag == 1} {InsertLog "cgidata= ([llength $cgidata]) $cgidata"}

       if {[llength $cgidata] == 0} then {return}

       if {[llength $cgidata] < 2} then {
           InsertLog "CGI data is too short: $cgidata"
       } else {
           if {[lindex $cgidata 0] != "Widget"} then {
               InsertLog "Invalid CGI data: $cgidata"
           } else {

              set w [first [lindex $cgidata 1]]
              set n [second [lindex $cgidata 1]]
              set vl [third [lindex $cgidata 1]]

               switch -glob $w {

                   RESET    {set started 0}
                   CLEAR    {EmptyLog}
                   ELOG     {PrintLog}
                   RELOAD   {}
                   VARIABLES {show_variables}
                   HELP     {Help $vl}

                   HIDE     {set ShowLogFlag 0}
                   SHOW     {set ShowLogFlag 1}
                   LOGON    {set LogFlag 1}
                   LOGOFF   {set LogFlag 0}

                   RANGE   -
                   TEXT    -
                   SETTING {do_Setting $n $vl}

                   UNIT    {do_Units $n}
                   GO      {do_stopgo $n}
                   RAMP    {do_Ramp $n}

                   REDISPLAY {UpdateModule}
                   UPDATE    {UpdateModule}

                   EXPERT   {$vl [third [lindex $cgidata 1]]}

                   MainLog {}
                   undefined {}
                   default {InsertLog "unexpected CGI data:  $cgidata"}
               }
           }
        }
        UpdateModule
     }
     UpdateModule

}   ;# end of if {$started == 0}




