
#    Note:  This is loaded into the namespace  MEISCI

# variable started 0

#Integrator.Current            : -a=0x00009000 -c7 -g=0
#Integrator.Current.Range      : -a=0x0000808b -d=8 -l=4 -b=3
#Integrator.Dose               : -a=0x00008088 -w16 -d8 -ud -mn -prw -l0 -b0 -i0 -zd 
#Integrator.Dose.Step          : -a=0x0000808b -d=8 -l=6 -b=1
#Integrator.Load               : -a0x00008000 -o0x0000008a -w16 -d8 -ud -mn -prw -l1 -b7 -i0 -zd -s0
#Integrator.PPS                : -a=0x0000808b -d=8 -l=2 -b=1
#Integrator.Percent            : -a=0x0000808c
#Integrator.Preset             : -a0x00008000 -o0x0000008e -w16 -d16 -ud -mn -prw -l0 -b0 -i%0000000000000000 -zb -s0
#Integrator.Range              : -a0x00008000 -o0x0000008b -w16 -d8 -ud -mn -pro -l4 -b3 -i0 -zd -s0
#Integrator.Readout            : -a0x00008000 -o0x0000008c -w16 -d16 -ud -mn -pro -l0 -b0 -i%0000000000000000 -zb -s0
#Integrator.Running            : -a0x00008000 -o0x0000008b -w16 -d8 -ud -mn -pro -l1 -b0 -i0 -zd -s0
#Integrator.Start              : -a0x00008000 -o0x0000008a -w16 -d8 -ud -mn -prw -l1 -b5 -i1 -zd -s0
#Integrator.Stop               : -a0x00008000 -o0x0000008a -w16 -d8 -ud -mn -prw -l1 -b6 -i1 -zd -s0

#xvme560   0x9000
#xvme240   0x8000



if {![info exists started]} {variable started 0}

if {$started != 0} {

     variable JS ""
     variable LogFlag

     if {$LogFlag == 1} {InsertLog "Last Updated: [clock format [clock seconds] -format "%B %d, %Y %H:%M:%S"]"}

    do_click
}

if {$started == 0} {           ;#  first time load code and define local variables
    set started 1

    set Color1(0) Red
    set Color1(1) Green
    set Color2(0) Cyan
    set Color2(1) #D4A017     ;# Orange Gold

    variable ShowLogFlag 0

    variable LogText ""
    variable LogFlag 0
    variable LogCount 0

    variable JS ""
    variable HelpTopic "undefined"

    variable STOP 1
    variable START 0
    variable RANGE 0
    variable Offset0 0
    variable Offset1 0
    variable Offset2 0
    variable PPS 0
    variable IPC 0
    variable CPC 0
    variable DOSE 0

    variable dosestep 600
    variable dosemax [expr $dosestep * 10000]
    variable doseunit C
    
    variable intrunning 0
    variable intpc 0


   proc  show_variables {} {
    variable STOP
    variable START
    variable RANGE
    variable PPS
    variable DOSE
    variable Offset0
    variable Offset1
    variable Offset2

  set registerList [list  Integrator.Dose Integrator.Dose.Step Integrator.Current Integrator.Current.Range Integrator.Percent Integrator.PPS]

       InsertLog "$STOP $START $RANGE $PPS $DOSE"
       InsertLog "$Offset0 $Offset1 $Offset2"

      foreach register $registerList {
         set rc [catch {ReadMEISRegister $register} m]
         InsertLog "$register => $rc $m"
      }
   }

   proc Help {subject} {
       variable HelpTopic

       switch $subject {

          Introduction -
          "Introduction to the System"  {set HelpTopic UsingMEISDataAcq}

          default {}
       }

       do_help
   }

    proc check-acquisition {m} {

        set going [ReadMEISRegister MEIS.Go]
        if {$going == 1} then {InsertLog "Error:  In Integrator: MEIS Data Acquisition Going - $m"}
    }
    
    proc check-acqrunning {} {
        variable intrunning
        set intrunning [ReadMEISRegister Integrator.Running]
    }
    
    proc do_reset {} {
    
        WriteMEISRegister Integrator.Reset 1
        WriteMEISRegister Integrator.Reset 0
    }
    
    proc do_convert {} {
      
        WriteMEISRegister Integrator.Convert 1   
    
    }

   proc do_Setting {reg vl} {        ;InsertLog "Setting $reg $vl"
      variable Offset0
      variable Offset1
      variable Offset2

#  w=0 Current (ro); w=1 Integrated (ro); w=2 Dose 

       switch $reg {
           rangeInput0   {set Offset0 $vl}
           rangeInput1   {set Offset1 $vl}
           rangeInput2   {set Offset2 $vl; do_Dose $vl}
           default {InsertLog "do_Setting: unexpected request $reg $vl"}
       }
   }
   
   proc do_MoveCup {vl} {   ;# this is disabled  - see meis/tcl/int.tcl
   
   
   }
   
   proc do_Dose {vl} {
      variable dosestep
      variable dosemax
      variable doseunit
   
      set value [ReadMEISRegister Integrator.Dose]
      scan $value %f%s val unit
      if {$unit != $doseunit} then {set val [expr $val*1000]}
      set val [first [split [expr $val+0.5] .]]
      if {$vl > $val && $vl < [expr $val+$dosestep]} then {set vl [expr $val+$dosestep]}
      if {$vl < $val && $vl > [expr $val-$dosestep]} then {set vl [expr $val-$dosestep]}
 
      set v "$val"
      WriteMEISRegister Integrator.Dose $v
      SetFooter "Integrator.Dose set to $v"  
   }

   proc do_Run {vl} {InsertLog "Run $vl"
      variable STOP
      variable START

       switch $vl {
         0 {
           check-acquisition "Attempting to Stop Integrator"
            do_MoveCup 0
            WriteMEISRegister Integrator.Stop 1
            WriteMEISRegister Integrator.Stop 0
            WriteMEISRegister Integrator.Stop 1
            set START 0
            set STOP 1
         }
         1 {
           check-acquisition "Attempting to Start Integrator"
            do_MoveCup 1
            WriteMEISRegister Integrator.Start 1
            WriteMEISRegister Integrator.Start 0
            WriteMEISRegister Integrator.Start 1
            set STOP 0
            set START 1
 
         }
         default {}
       }
 
   }

    proc UpdateCurrent {} {
       variable Offset0
       variable CPC
        variable RANGE

       set vl [ReadMEISRegister Integrator.Current]
       set Offset0 $vl

       set vl [ReadMEISRegister Integrator.Current.Range]
       set RANGE $vl
       set max [scan-int $vl]
       set val [expr ([scan-int $Offset0] * $max) / 4095]
       set CPC $val
    }

    proc UpdatePercent {} {
       variable Offset1
       variable IPC

       set vl [ReadMEISRegister Integrator.Percent]
       set val [scan-int $vl]
       set IPC $val

       if {$vl > 100} {set vl 100}
       set val [scan-int $vl]
       set Offset1 $val

    }

    proc UpdateDose {} {
      variable dosestep
      variable dosemax
      variable doseunit
      variable Offset2

       set vl [ReadMEISRegister Integrator.Dose]
       scan $vl %f%s val unit
#       if {$unit != $doseunit} then {set val [expr $val*1000]}
       set val [format %0.3f $val]
       set Offset2 $val
    }

    proc UpdateDoseStep {} {
      variable dosestep
      variable dosemax
      variable doseunit

      set vl [ReadMEISRegister Integrator.Dose.Step]

      set dosestep [first $vl]
      set dosemax [expr $dosestep * 10000]
      set doseunit [second $vl]

    }

    proc UpdatePPS {} {
        variable PPS

        set PPS [ReadMEISRegister Integrator.PPS]
    }

    proc UpdateModule {} {

       UpdateCurrent   ;#  register Integrator.Current
       UpdatePercent   ;#  register Integrator.Percent
       UpdateDoseStep  ;#  register Integrator.Dose.Step
       UpdatePPS       ;#  register Integrator.PPS
       
       do_redisplay
 
   }
   
   proc do_redisplay {} {
       variable intpc
       variable intrunning
       
       set intrunning [ReadMEISRegister Integrator.Running]
       if {$intrunning == 0} {WriteMEISRegister MEIS.Go 0}

       UpdateDose   ;# register Integrator.Dose
       
   }

   proc  do_click {}  {
       variable cgidata
       variable started
       variable LogFlag
       variable ShowLogFlag

       set cgidata [ncgi::nvlist]

       if {$LogFlag == 1} {InsertLog "cgidata= ([llength $cgidata]) $cgidata"}

       if {[llength $cgidata] == 0} then {return}

       if {[llength $cgidata] < 2} then {
           InsertLog "CGI data is too short: $cgidata"
       } else {
           if {[lindex $cgidata 0] != "Widget"} then {
               InsertLog "Invalid CGI data: $cgidata"
           } else {

              set w [first [lindex $cgidata 1]]
              set n [second [lindex $cgidata 1]]
              set vl [third [lindex $cgidata 1]]

               switch -glob $w {

                   RESET    {set started 0}
                   CLEAR    {EmptyLog}
                   ELOG     {PrintLog}
                   RELOAD   {}
                   VARIABLES {show_variables}
                   HELP     {Help $vl}

                   HIDE     {set ShowLogFlag 0}
                   SHOW     {set ShowLogFlag 1}
                   LOGON    {set LogFlag 1}
                   LOGOFF   {set LogFlag 0}

                   RANGE   -
                   TEXT    -
                   SETTING {do_Setting $n $vl}

                   RUN     {do_Run $n}
                   DOSE    {do_Dose}
                   

                   REDISPLAY {}
                   UPDATE    {}

                   EXPERT   {$n $vl}

                   MainLog {}
                   undefined {}
                   default {InsertLog "unexpected CGI data:  $cgidata"}
               }
           }
        }
        UpdateModule
     }
     UpdateModule

}   ;# end of if {$started == 0}




