

#    experiment run control for MEIS - taken from sys.tcl

global MIDAS_DataAcquisitionControlServer_Verbose
set MIDAS_DataAcquisitionControlServer_Verbose 0

variable Options

variable SpecBase [file join $env(MIDASBASE) SpecData $::DataAcquisition::PROJECT]
variable SpecPath


set z [catch {variable Modules $::VMEConfig::ModuleMenu} m]
if {$z != 0} {variable Modules [list CAEN_V785N SAC ERDA]}

variable Supported_Modules {}
foreach item $Modules {
     if {$item == "CAEN_V785"} {lappend Supported_Modules V785; continue}
     if {$item == "CAEN_V785N"} {lappend Supported_Modules V785N; continue}
     if {$item == "SAC"} {lappend Supported_Modules $item; continue}
     if {$item == "VData"} {lappend Supported_Modules $item; continue}
     if {$item == "ERDA"} {lappend Supported_Modules $item; continue}
}

puts "ConfigBase = $::VMEConfig::ConfigBase"
puts "Supported_Modules = $Supported_Modules"


proc load_Options {} {LoadOptions}
proc update_Options {} {SaveOptions}


   proc Set_Options {} {
      variable Options
      variable Supported_Modules

      LoadOptions

      if {![info exists Options(Rate.channels)]} {set Options(Rate.channels) 512}
      if {![info exists Options(Stat.channels)]} {set Options(Stat.channels) 512}
      if {![info exists Options(Stat.shift)]} {set Options(Stat.shift) 6}
      if {![info exists Options(Stat.offset)]} {set Options(Stat.offset) 0}

   if {![info exists Options(VME.settings)]} {set Options(VME.settings) undefined}
   if {![info exists Options(Usepatterngen)]} {set Options(Usepatterngen) 0}

   if {![info exists Options(UseSAC)]} {set Options(UseSAC) 0}
   if {![info exists Options(Use1495)]} {set Options(Use1495) 0}
   if {![info exists Options(Use1495DPP)]} {set Options(Use1495DPP) 0}

      foreach module $Supported_Modules {
         set Options(Include.$module)  1
         set Options(Histogram.$module) 1
      }

      SaveOptions
   }

   proc Dump_Raw_Data {} {
      variable Supported_Modules

#      foreach module $Supported_Modules {append info [Dump_Raw_[set module]_Data]}

    set info "Raw data dump for MEIS VME modules\n"

      set maxmodule 0
      foreach module $Supported_Modules {
         for {set i 1} {$i <= 16} {incr i} {
            set z [catch {ReadVMERegisterAttr [string tolower $module]#$i.VMEStatus} m]
            if {$z!=0} then {break}
            incr maxmodule
         }
      }

    for {set i 1} {$i <= $maxmodule} {incr i} {

       append info "VXXXX#$i\n"

        DataAcqCommand 19 "Getting Raw Data" "1 $i" 2
        set data [ReadNetVar NetVar.EXEC.Response]

        foreach word $data {
           if {[string range $word 0 1] != "0x"} then {set word 0x$word}
           if {$word == 0xaa5555aa} then {
               append info "$word - end of dump\n"
               break
           }
           if {$word == 0x55aaaa55} then {
               append info "$word - end of event\n"
               continue
           }
           append info "$word "
        }

        append info "\n"
     }


       append info "VME\n"

        DataAcqCommand 19 "Getting Raw Data" "5 0" 2
        set data [ReadNetVar NetVar.EXEC.Response]

        foreach word $data {
           if {[string range $word 0 1] != "0x"} then {set word 0x$word}
           if {$word == 0xaa5555aa} then {
               append info "$word - end of dump\n"
               break
           }
           if {$word == 0x55aaaa55} then {
               append info "$word - end of event\n"
               continue
           }
           append info "$word "
        }

        append info "\n"

     return $info

   }

   proc Stop_Electronics {} {
       variable Options 
       variable Supported_Modules

       if {$Options(UseSAC) != 0} {WriteVMERegister sac#1.STR 0}   ;#  disable acquisition

       foreach m $Supported_Modules {[set m]_DAQ-halt}

     after 2000   ;#  allow DA program to flush data 

   }

   proc Go_Electronics {} {
      variable Options 
      variable Supported_Modules

puts "starting Go_Electronics"

puts "UseSAC=$Options(UseSAC)"

       foreach m $Supported_Modules {[set m]_DAQ-go}

       if {$Options(UseSAC) != 0} {WriteVMERegister sac#1.STR 9;  WriteVMERegisters sac#1.ARM 0}

puts "Go_Electronics complete"
   }
 
   proc Reset_Electronics {} {    ;#   used by RESET
      variable Supported_Modules

      foreach m $Supported_Modules {[set m]_DAQ-reset}
   }

   proc Setup_Electronics {} {    ;#   used by SETUP
       variable Supported_Modules
       variable Options

       define_VME_addresses   ;#  setup VME 

       foreach m $Supported_Modules {[set m]_DAQ-setup}

   }

   proc define_VME_addresses {} {
       variable Supported_Modules
       variable Options

#    load hardware configuration into readout software

       ClearConfig
 
       foreach module $Supported_Modules {
          if {$Options(Include.$module) == 1} then {define_[set module]_config}
       }

       DataAcqCommand 4 "print config" 0 2

   }



    proc TerminateExec {} {

#        unmap_spectra
#        DataAcqCommand 22 "Terminating DataAcq program" 0 2

       define_VME_addresses

    }

    proc LoadExec {} {

       set z [catch {eval exec nice [file join /MIDAS Data_Acq bin Linux64 CAENExecV8] &} m]
       if {$z != 0} {puts "RunControl LoadExec returned z=$z and $m"}

    }


    foreach module $Supported_Modules {

        set f [file join $env(MIDASBASE) TclHttpd Html MEIS RunControl sys.$module.tcl]
        set z [catch {source $f} m]
        if {$z == 0} {puts "RunControl loaded $f"} else {puts "RunControl returned z=$z and $m"}
    }

#    set f [file join $env(MIDASBASE) TclHttpd Html MEIS RunControl sys.ERDA.tcl]
#    set z [catch {source $f} m]
#    if {$z == 0} {puts "RunControl loaded $f"} else {puts "RunControl returned z=$z and $m"}

    make-UIregisters

    Set_Options

    foreach module $Supported_Modules {Set_[set module]_Options}

#    Set_ERDA_Options







