
#    MIDAS Httpd Service to provide VME Access

package require MIDASDomainHandler     ;#  load default support procedures

#     Note that the method of VME access (e.g. via the CAEN V1718 USB2 interface) is defined in the server startup file

global MIDAS_VMEAccessServer_Exists
set MIDAS_VMEAccessServer_Exists 1

global MIDAS_VMEAccessServer_Verbose
set MIDAS_VMEAccessServer_Verbose 0

global MIDAS_SOAPDomainHandler_Verbose
set MIDAS_SOAPDomainHandler_Verbose 0

namespace eval VMEAccessServer {
    variable DBRootName; if {![info exists DBRootName]} {set DBRootName [file join EXPERIMENTS VME]}
    variable DBNodeBase; if {![info exists DBNodeBase]} {set DBNodeBase undefined}
    variable DBNodeName; if {![info exists DBNodeName]} {set DBNodeName [exec uname -n]}
}

if {![info exists ::VME::TestMode]} {namespace eval VME {set TestMode 0}}
if {$::VME::TestMode != 0} {
    global VMEData 
    set VMEData(BASE) 1
}


# Use namespaces to isolate your methods
namespace eval urn:VMEAccessServer {

#    package require RegisterAccessServer       ;#  load common procedures

   global env
   set z [catch {source [file join $env(MIDASBASE) TclHttpd tcl RegisterAccessServer.tcl]} m]
   if {$z != 0} {puts "RegisterAccessServer.tcl > $m"}

#    set up the Instrinsic registers

    set RegisterBase 0

    set HasBeenConfigured 0
    set HasBeenRestored 0

    set Address ""
    set Data 0
    set Status 0
    set Debug 0

    set RegisterIndex 1
    
    set RegisterNames(HasBeenConfigured) $RegisterIndex
    set RegisterNames(NetVar_HasBeenConfigured) $RegisterIndex
    set R[set RegisterIndex](class) vmeConfiguredState
    incr RegisterIndex

    set RegisterNames(HasBeenRestored) $RegisterIndex
    set RegisterNames(NetVar_HasBeenRestored) $RegisterIndex
    set R[set RegisterIndex](class) vmeRestoredState
    incr RegisterIndex

    set RegisterNames(VME.Address) $RegisterIndex
    set R[set RegisterIndex](class) vmeAddr
    incr RegisterIndex

    set RegisterNames(VME.Execute) $RegisterIndex
    set R[set RegisterIndex](class) vmeExec
    incr RegisterIndex

    set RegisterNames(VME.Data) $RegisterIndex
    set R[set RegisterIndex](class) vmeData
    incr RegisterIndex

    set RegisterNames(VME.Status) $RegisterIndex
    set R[set RegisterIndex](class) vmeStatus
    incr RegisterIndex

    set RegisterNames(VME.Debug) $RegisterIndex
    set R[set RegisterIndex](class) vmeDebug
    incr RegisterIndex

    set RegisterBase $RegisterIndex

#    xVME register attributes

   set xVME_options {
         {a.arg 0    "VME address base"}
         {o.arg 0    "VME address offset"}
         {w.arg 32   "address width"}
         {d.arg 32   "data width"}
         {u.arg "d"  "use"}
         {m.arg "n"  "mode or VME address modifier"}
         {p.arg "rw" "access permission"}
         {b.arg 0    "field width (bits)"}
         {l.arg 0    "field length (bits)"}
         {z.arg "x"  "data format returned"}
         {s.arg "0"  "return status"}
         {i.arg 0xa55a  "initial value"}
   }

   set xVME_Updateoptions {
         {a.arg 0xa55a    "VME address base"}
         {o.arg 0xa55a    "VME address offset"}
         {w.arg 0xa55a    "address width"}
         {d.arg 0xa55a    "data width"}
         {u.arg 0xa55a    "use"}
         {m.arg 0xa55a    "mode or VME address modifier"}
         {p.arg 0xa55a    "access permission"}
         {b.arg 0xa55a    "field width (bits)"}
         {l.arg 0xa55a    "field length (bits)"}
         {z.arg 0xa55a    "data format returned"}
         {s.arg 0xa55a    "return status"}
         {i.arg 0xa55a    "initial value"}
    }

    set xVME_Usage ": xVME  -a ? -o ? -w 16|24|32 -d 8|16|32 -u d|p -m n|s|? -p ro|wo|rw -b ? -l ? -z w|b|d -s 0|1 -i ?"

#    bVME register attributes

   set bVME_options {
         {a.arg 0    "VME address base"}
         {o.arg 0    "VME address offset"}
         {w.arg 32   "address width"}
         {d.arg 32   "data width"}
         {u.arg "d"  "use"}
         {m.arg "n"  "mode or VME address modifier"}
         {p.arg "rw" "access permission"}
         {z.arg "x"    "data format returned"}
         {l.arg "1"  "length in data items"}
         {i.arg 0xa55a  "initial value"}
   }

   set bVME_Updateoptions {
         {a.arg 0xa55a    "VME address base"}
         {o.arg 0xa55a    "VME address offset"}
         {w.arg 0xa55a    "address width"}
         {d.arg 0xa55a    "data width"}
         {u.arg 0xa55a    "use"}
         {m.arg 0xa55a    "mode or VME address modifier"}
         {p.arg 0xa55a    "access permission"}
         {z.arg 0xa55a    "data format returned"}
         {l.arg 0xa55a    "length in data items"}
         {i.arg 0xa55a    "initial value"}
    }

    set bVME_Usage ": xVME  -a ? -o ? -w 16|24|32 -d 8|16|32 -u d|p -m n|s|? -p ro|wo|rw -z w|b|d -l ? -i ?"


#    aVME register attributes

   set aVME_options {
         {a.arg 0    "VME address base"}
         {o.arg 0    "VME address offset"}
         {w.arg 32   "address width"}
         {d.arg 32   "data width"}
         {m.arg "n"  "mode or VME address modifier"}
         {p.arg "rw" "access permission"}
         {q.arg ""   "prefix"}
         {u.arg ""   "unit"}
         {r.arg "x0:0,0xffff:65535"  "range=min (hex):(real), max (hex):(real)"}
         {s.arg "0"  "return status"}
         {i.arg 0xa55a  "initial value"}
   }

   set aVME_Updateoptions {
         {a.arg 0xa55a    "VME address base"}
         {o.arg 0xa55a    "VME address offset"}
         {w.arg 0xa55a   "address width"}
         {d.arg 0xa55a   "data width"}
         {m.arg 0xa55a  "mode or VME address modifier"}
         {p.arg 0xa55a "access permission"}
         {q.arg 0xa55a   "prefix"}
         {u.arg 0xa55a   "unit"}
         {r.arg 0xa55a  "range=min (hex):(real), max (hex):(real)"}
         {s.arg 0xa55a  "return status"}
         {i.arg 0xa55a  "initial value"}
    }

    set aVME_Usage ": aVME  -a ? -o ? -w 16|24|32 -d 16|32 -m n|s|? -p ro|wo|rw -q ? -u ? -r ?:?,?:? -s 0|1 -i ?"

    variable QUAL
    set QUAL {
        {a  1.0e-18}                                                ;#  /* ato   */
        {f  1.0e-15}                                                ;#   /* femto */
        {p  1.0e-12}                                                ;#  /* pico  */
        {n  1.0e-9 }                                               ;#   /* nano  */
        {u  1.0e-6 }                                                ;#  /* micro */
        {m  1.0e-3 }                                              ;#    /* milli */
        {c  0.01   }                                            ;#       /* centi */
        {d  0.1    }                                                ;#  /* deci  */
        {D  10.0   }                                              ;#    /* deca  */
        {h  100.0  }                                               ;#   /* hecto */
        {K  1.0e3  }                                                ;#   /* kilo  */
        {k  1.0e3  }                                                ;#   /* kilo  */
        {M  1.0e6  }                                                ;#   /* mega  */
        {G  1.0e9  }                                                 ;#  /* giga  */
        {T  1.0e12 }                                                ;#   /* tera  */
        {P  1.0e15 }                                             ;#     /* peta  */
        {E  1.0e18 }                                               ;#    /* exa   */
        {*  1.0  }}                                                ;#   /* default  */

    proc scan-num s {    ;# remove after common.tcl updated
       if {[string first "0x" $s]==0} then {
         set z [scan [string range $s 2 end] %x zzz]
       } else {
         set z [scan $s %f zzz]
       }
       if {$z==1} then {return $zzz} else {return "???"}
    }

    proc stor str {      ;#   string str to real conversion     ie  1.2 mV => 0.0012
       variable QUAL

puts "stor: $str"
        scan $str "%f%n" value n
        set q [string range $str $n $n]    ;# extract the qualifier
        set vl $value
        foreach item $QUAL {
           if {[first $item] == $q} {set vl [expr $value * [second $item]]; break}
        }
puts "stor: $item value=$vl"
        return $vl
    }


    proc rtos {value q u} {          ;#  real value to string conversion   ie 0.0012 => 1.2mV
        variable QUAL

        foreach item $QUAL {
           if {[first $item] == $q} {set value [expr $value / [second $item]]}
        }
        if {$q == "*"} {
            set vl [format "%.3f%s" [expr 1.0 * $value] $u]
        } else {
            set vl [format "%.3f%s%s" [expr 1.0 * $value] $q $u]
        }
puts "rtos: value $q $u => $vl"
        return $vl
    }

    proc dtoa {value M C} {       ;#  digital to analogue conversion    y = (m * x) + c
puts "dtoa: $value $M $C"
       return [expr ($M * $value) + $C]
    }

    proc atod {value M C} {       ;#  analogue to digital conversion    x = (y - c) / m
puts "atod: $value $M $C"
      set vl [expr round(($value - $C) / $M)]
      return $vl
    }

    proc AccessServerVerbose {option} {
        global MIDAS_VMEAccessServer_Verbose
        global MIDAS_RegisterAccessServer_Verbose
        global MIDAS_SOAPDomainHandler_Verbose
        set MIDAS_VMEAccessServer_Verbose $option
        set MIDAS_RegisterAccessServer_Verbose $option
        set MIDAS_SOAPDomainHandler_Verbose $option
        return [rpcvar::rpcvar int 0]
    }

    proc DefineRegister {Register Class Attributes} {
        global MIDAS_VMEAccessServer_Verbose
        upvar RegisterIndex RegisterIndex
        upvar RegisterNames RegisterNames

        if {$MIDAS_VMEAccessServer_Verbose} \
             {puts "called method VMEAccessServer::DefineRegister with parameters $Register $Class $Attributes"}
             
#    Check Register name is not already in use

        if {[info exists RegisterNames($Register)]} then {
             return -code error -errorcode 0x20006 -errorinfo "Register name already in use"
        }

#    Check Class is valid

        switch $Class {
            xVME -
            aVME -
            bVME  {}

            default {
               return -code error -errorcode 0x20007 -errorinfo "Class name is invalid or not supported"
            }
        }

#    create register

        switch $Class {
            xVME  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar xVME_options options

                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }
                set R(am) 0x09

                set R(LastRead)  0x0
                set R(LastWrite) 0x0

                incr RegisterIndex
            }
            aVME  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar aVME_options options

                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }
                set R(am) 0x09

                set R(LastRead)  0x0
                set R(LastWrite) 0x0
                set R(Dmin) 0
                set R(Dmax) 0xffff
                set R(Amin) 0.0
                set R(Amax) 65535.0
                set R(M) 1.0
                set R(C) 0
                set R(prefix) " "
                set R(unit) ""

                incr RegisterIndex
            }
            bVME  {

                set RegisterNames($Register) $RegisterIndex

                upvar R$RegisterIndex R
                set R(class) $Class

                upvar bVME_options options

                foreach item $options {
                    set opt [first [split [first $item] .]]
                    set R($opt) [second $item]
                }
                set R(am) 0x0b   ;# BLT
                
                incr RegisterIndex
            }
            default {
                return -code error -errorcode 0x20007 -errorinfo "Class name is invalid or not supported"
            }
        }

#    apply attributes

        return [do_WriteRegisterAttributes $Register $Attributes]
    }


    proc do_ReadRegisterAttributes {Register} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        switch $R(class) {
            xVME {
                 set attributes "-a$R(a) -o $R(o) -w $R(w) -d $R(d) -m $R(m) -p $R(p) -l $R(l) -b $R(b) -z $R(z) -s $R(s)"
                 if {$R(i) != 0xa55a} {append attributes " -i$R(i)"}
            }
            aVME {
                 set attributes "-a$R(a) -o $R(o) -w $R(w) -d $R(d) -m $R(m) -p $R(p) -q $R(q) -u $R(u) -r $R(r) -s $R(s)"
                 if {$R(i) != 0xa55a} {append attributes " -i$R(i)"}
            }
            bVME {
                 set attributes "-a$R(a) -o $R(o) -w $R(w) -d $R(d) -m $R(m) -p $R(p) -z $R(z) -l $R(l)"
                 if {$R(i) != 0xa55a} {append attributes " -i $R(i)"}
            }
            default {
                return -code error -errorcode 0x10004 -errorinfo "Invalid request"
            }
        }

        return $attributes
    }

    proc do_WriteRegisterAttributes {Register Attributes} {
       variable QUAL

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames

        set Index $RegisterNames($Register)
        upvar 2 R$Index R

        if {$MIDAS_VMEAccessServer_Verbose} \
              {puts "VME WriteAttributes: $Register $Attributes $R(class)"}

        switch $R(class) {

            xVME  {

                upvar 2 xVME_Updateoptions options
                upvar 2 xVME_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        a  {}
                        o  {}
                        w  {if {$params(w) != 16 && $params(w) != 24 && $params(w) != 32} {set rc -1}}
                        d  {if {$params(d) != 8 && $params(d) != 16 && $params(d) != 32} {set rc -1}}
                        u  {if {$params(u) != "d" && $params(u) != "p"} {set rc -1}}
                        m  {
                           if {[scan-num $params(m)] == "???"} then {
                               if {$params(m) != "n" && $params(m) != "s"} {set rc -1}
                           } else {
                              switch [format 0x%02x $params(m)] {
                                 0x29 -
                                 0x2d -
                                 0x39 -
                                 0x3d -
                                 0x3a -
                                 0x3e -
                                 0x09 -
                                 0x0d -
                                 0x0a -
                                 0x0e  {set params(m) [format 0x%02x $params(m)]}
                              }
                           }
                        }
                        p  {if {$params(p) != "rw" && $params(p) != "ro" && $params(p) != "wo"} {set rc -1}}
                        l  {if {$params(l) < 0 || $params(l) > 32} {set rc -1}}
                        b  {if {$params(b) < 0 || $params(b) > 32} {set rc -1}}
                        z  {if {$params(z) != "d" && $params(z) != "x" && $params(z) != "b"} {set rc -1}}
                        i  {}
                        s  {if {$params(s) != 0 && $params(s) != 1} {set rc -1}}
                        default {set rc -1}
                    }
                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
                }
                if {[scan-num $R(m)] == "???"} {
                    switch [set R(u)][set R(m)][set R(w)] {
                        dn16 {set R(am) 0x29}
                        ds16 {set R(am) 0x2d}
                        pn16 {set R(am) 0x29}
                        ps16 {set R(am) 0x2d}
                        dn24 {set R(am) 0x39}
                        ds24 {set R(am) 0x3d}
                        pn24 {set R(am) 0x3a}
                        ps24 {set R(am) 0x3e}
                        dn32 {set R(am) 0x09}
                        ds32 {set R(am) 0x0d}
                        pn32 {set R(am) 0x0a}
                        ps32 {set R(am) 0x0e}
                    }
                } else {
                    set R(am) $R(m)
                }
            }
            aVME  {

                upvar 2 aVME_Updateoptions options
                upvar 2 aVME_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        a  {}
                        o  {}
                        w  {if {$params(w) != 16 && $params(w) != 24 && $params(w) != 32} {set rc -1}}
                        d  {if {$params(d) != 16 && $params(d) != 32} {set rc -1}}
                        m  {
                           if {[scan-num $params(m)] == "???"} then {
                               if {$params(m) != "n" && $params(m) != "s"} {set rc -1}
                           } else {
                              switch [format 0x%02x $params(m)] {
                                 0x29 -
                                 0x2d -
                                 0x39 -
                                 0x3d -
                                 0x3a -
                                 0x3e -
                                 0x09 -
                                 0x0d -
                                 0x0a -
                                 0x0e  {set params(m) [format 0x%02x $params(m)]}
                              }
                           }
                        }
                        p  {if {$params(p) != "rw" && $params(p) != "ro" && $params(p) != "wo"} {set rc -1}}
                        q  {
                           foreach qual $QUAL {
                              if {$params(p) == [first $qual]} {set R(prefix) [second $qual]; break}
                           }
                        }
                        u  {}
                        r  {
                           set Min [first [split $params(r) ,]]    ;#ie 0:0A
                           set Max [second [split $params(r) ,]]   ;#ie 0xfff:1.5mA
                           set R(Dmin) [first [split $Min :]]   ;# digital min 0
                           set R(Amin) [second [split $Min :]]  ;# analogue min 0A
                           set R(Dmax) [first [split $Max :]]   ;# digital max 0xfff
                           set R(Amax) [first [split $Max :]]   ;# analogue max 1.5mA

                           set field $R(Amin)                ;# 0A
                           scan $field "%f%n" value n   
                           set R(Umin) [string range $field $n end]   ;# A

                           set field $R(Amax)                ;# 1.5mA
                           scan $field "%f%n" value length  
                           set R(Umax) [string range $field $length end]   ;# mA

                           if {$R(Dmin) > $R(Dmax)} {
                              set x $R(Amax)
                              set R(Amax) $R(Amin)
                              set R(Amin) $x
                              set x $R(Dmax)
                              set R(Dmax) $R(Dmin)
                              set R(Dmin) $x
                              set x $R(Umax)
                              set R(Umax) $R(Umin)
                              set R(Umin) $x
                           }
                           set rMin [stor $R(Amin)]     ;#   0A => 0.0
                           set rMax [stor $R(Amax)]     ;#   1.5mA => 0.0015
                           set R(M) [expr ($rMax - $rMin) / ($R(Dmax) - $R(Dmin))]
                           set R(C) [expr $R(Dmin) - ($R(M) * $rMin)]

        if {$MIDAS_VMEAccessServer_Verbose} \
              {puts "VME WriteAttributes aVME $Min $Max $R(Dmin) $R(Dmax) $R(Amin) $R(Amax) $rMin $rMax $R(M) $R(C)"}
                        }
                        i  {}
                        s  {if {$params(s) != 0 && $params(s) != 1} {set rc -1}}
                        default {set rc -1}
                    }
                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
#                    if {$m == "i"} {do_WriteRegister $Register [set R(i)]}
                }
                if {[scan-num $R(m)] == "???"} {
                    switch [set R(u)][set R(m)][set R(w)] {
                        dn16 {set R(am) 0x29}
                        ds16 {set R(am) 0x2d}
                        pn16 {set R(am) 0x29}
                        ps16 {set R(am) 0x2d}
                        dn24 {set R(am) 0x39}
                        ds24 {set R(am) 0x3d}
                        pn24 {set R(am) 0x3a}
                        ps24 {set R(am) 0x3e}
                        dn32 {set R(am) 0x09}
                        ds32 {set R(am) 0x0d}
                        pn32 {set R(am) 0x0a}
                        ps32 {set R(am) 0x0e}
                    }
                } else {
                    set R(am) $R(m)
                }
            }
            bVME  {

                upvar 2 bVME_Updateoptions options
                upvar 2 bVME_Usage usage

                array set params [::cmdline::getKnownOptions Attributes $options $usage]
                
                foreach m [array names params] {
                    set rc 0
                    switch $m {
                        a  {}
                        o  {}
                        w  {if {$params(w) != 16 && $params(w) != 24 && $params(w) != 32} {set rc -1}}
                        d  {if {$params(d) != 8 && $params(d) != 16 && $params(d) != 32} {set rc -1}}
                        u  {if {$params(u) != "d" && $params(u) != "p"} {set rc -1}}
                        m  {
                           if {[scan-num $params(m)] == "???"} then {
                               if {$params(m) != "n" && $params(m) != "s"} {set rc -1}
                           } else {
                              switch [format 0x%02x $params(m)] {
                                 0x38 -
                                 0x3b -
                                 0x3c -
                                 0x3f -
                                 0x08 -
                                 0x0b -
                                 0x0c -
                                 0x0f  {set params(m) [format 0x%02x $params(m)]}
                              }
                           }
                        }
                        p  {if {$params(p) != "rw" && $params(p) != "ro" && $params(p) != "wo"} {set rc -1}}
                        z  {if {$params(z) != "d" && $params(z) != "x" && $params(z) != "b"} {set rc -1}}
                        l  {if {$params(l) < 1 || $params(l) > 1024} {set rc -1}}
                        i  {}
                        default {set rc -1}
                    }
                    if {$rc == 0 && $params($m) != 0xa55a} {set R($m) $params($m)}
                }
                if {[scan-num $R(m)] == "???"} {
                    switch [set R(u)][set R(m)][set R(w)] {
                        dn24 {set R(am) 0x3b}
                        ds24 {set R(am) 0x3f}
                        dn32 {set R(am) 0x0b}
                        ds32 {set R(am) 0x0f}
                    }
                } else {
                    set R(am) $R(m)
                }
            }
            default {
                return -code error -errorcode 0x10004 -errorinfo "Invalid request"
            }
        }
 
        return 0
    }

    proc do_WriteRegister {Register Value} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames
        upvar 2 HasBeenConfigured HasBeenConfigured
        upvar 2 HasBeenRestored HasBeenRestored
        upvar 2 Debug Debug
        upvar 2 Address Address

        set Index $RegisterNames($Register)
        upvar 2 R$Index R
        if {![info exists R]} {upvar 3 R$Index R}

       if {$MIDAS_VMEAccessServer_Verbose} \
              {puts "VME WriteRegister: $Register $Value $R(class)"}

        switch $R(class) {
            xVME  {
                if {$R(p) == "ro"} {
                    return -code error -errorcode 0x20004 -errorinfo "Register is not writeable"
                }

                if {($R(b) != 0) || ($R(l) != 0) } then {    ;#   need to get the current value

                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) u=$R(u) m=$R(m)"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(am)

                    if {$::VME::TestMode != 0} {
                      global VMEData
                      if {![info exists VMEData($Register)]} {set VMEData($Register) 0x0003dead}
                      set rc [list 0 [set VMEData($Register)]]
                    } else {
                        set rc [rdVME -a $Addr -w $DataWidth -m $AddrMod]
                    }

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                    if {[first $rc] != 0} {return -code error -errorcode $rc -errorinfo "VME access error"}
                    set data [second $rc]

#    data now contains data last written or data read back
#    it is raw - not masked or shifted
#    mask only those bits not to be changed by this write
#    then mask out those bits to be changed by this write
#    and OR in this write data

                    set data [expr $data  & ~(~(~0 << $R(l)) << $R(b))]
                    set vl [expr (($Value & ~(~0 << $R(l))) << $R(b)) | $data]

                } else {set vl $Value}

                if {$MIDAS_VMEAccessServer_Verbose} \
                    {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(m) data=$vl"}

                set Addr [expr $R(a) + $R(o)]
                set DataWidth $R(d)
                set AddrMod $R(am)

                    if {$::VME::TestMode != 0} {
                      global VMEData
                      set VMEData($Register) $vl
                      set rc [list 0 [set VMEData($Register)]]
                    } else {
                        set rc [wtVME -a $Addr -w $DataWidth -m $AddrMod -d $vl]
                    }

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

                set R(LastWrite) $Value
            }
            aVME {
                if {$R(p) == "ro"} {
                    return -code error -errorcode 0x20004 -errorinfo "Register is not writeable"
                }

                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) $R(Dmin) $R(Dmax) $R(Amin) $R(Amax) M=$R(M) C=$R(C) data=$Value"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(am)

                    set vl [atod [stor $Value] $R(M) $R(C)]  ;# convert from analogue to digital

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME Write: $Value $R(C) $R(M) $vl"}

                    if {$::VME::TestMode != 0} {
                      global VMEData
                      set VMEData($Register) $vl
                      set rc [list 0 [set VMEData($Register)]]
                    } else {
                        set rc [wtVME -a $Addr -w $DataWidth -m $AddrMod -d $vl]
                    }

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                    if {[first $rc] != 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

                set R(LastWrite) $vl
            }
            bVME  {
                if {$R(p) == "ro"} {
                    return -code error -errorcode 0x20004 -errorinfo "Register is not writeable"
                }
                if {$MIDAS_VMEAccessServer_Verbose} \
                    {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(m) data=$vl"}

                set Addr [expr $R(a) + $R(o)]
                set DataWidth $R(d)
                set AddrMod $R(am)

                set rc [wtblkVME -a $Addr -w $DataWidth -m $AddrMod -d $vl]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

            }
            vmeAddr {
                set Address "$Value"
            }
            vmeExec  {
               if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME request $Address"}

                set rc [wtVME -a $Addr -w $DataWidth -m $AddrMod -d $Value]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}
            }
            vmeConfiguredState {
                set HasBeenConfigured $Value
            }
            vmeRestoredState {
                set HasBeenRestored $Value
            }
            vmeDebug {
                set Debug $Value
            }
            default {
                return -code error -errorcode 0x10004 -errorinfo "Invalid request"
            }
        }
 
        return 0
    }

    proc do_ReadRegister {Register} {

        global MIDAS_VMEAccessServer_Verbose
        upvar 2 RegisterNames RegisterNames
        if {![info exists RegisterNames]} {upvar 3 RegisterNames RegisterNames}
        upvar 2 HasBeenConfigured HasBeenConfigured
        if {![info exists HasBeenConfigured]} {upvar 3 HasBeenConfigured HasBeenConfigured}
        upvar 2 HasBeenRestored HasBeenRestored
        if {![info exists HasBeenRestored]} {upvar 3 HasBeenRestored HasBeenRestored}
        upvar 2 Debug Debug
        if {![info exists Debug]} {upvar 3 Debug Debug}
        upvar 2 Address Address
        if {![info exists Address]} {upvar 3 Address Address}
        upvar 2 Data Data
        if {![info exists Data]} {upvar 3 Data Data}
        upvar 2 Status Status
        if {![info exists Status]} {upvar 3 Status Status}

        set Index $RegisterNames($Register)
        upvar 2 R$Index R
        if {![info exists R]} {upvar 3 R$Index R}

       if {$MIDAS_VMEAccessServer_Verbose} \
              {puts "VME ReadRegister: $Register $R(class)"}

        switch $R(class) {
            xVME  {
                if {$R(p) == "wo"} then {
                    set Value $R(LastWrite)
                } else {
                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) u=$R(u) m=$R(m)"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(am)

                    if {$::VME::TestMode != 0} {
                      global VMEData
                      if {![info exists VMEData($Register)]} {set VMEData($Register) 0x0003dead}
                      set rc [list 0 [set VMEData($Register)]]
                    } else {
                        set rc [rdVME -a $Addr -w $DataWidth -m $AddrMod]
                    }

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                    if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

                    set Value [second $rc]; set R(LastRead) $Value

                    if {($R(b) != 0) || ($R(l) != 0)} then {set Value [expr ($Value >> $R(b)) & ~(~0 << $R(l))]}

                    switch $R(z) {
                        x   {set Value [format "0x%0*x" [expr $R(d)/4] $Value]}
                        d   {set Value [format "%d" $Value]}
                        b   {}
                    }
                }
            }
            aVME  {
                if {$R(p) == "wo"} then {
                    set Value $R(LastWrite)
                } else {
                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(m) q=$R(q) u=$R(u) r=$R(r)"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(am)

                    if {$::VME::TestMode != 0} {
                      global VMEData
                      if {![info exists VMEData($Register)]} {set VMEData($Register) 0x0003dead}
                      set rc [list 0 [set VMEData($Register)]]
                    } else {
                        set rc [rdVME -a $Addr -w $DataWidth -m $AddrMod]
                    }

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                    if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

                    set Value [second $rc]; set R(LastRead) $Value

                }

                if {$Value > $R(Dmax)} {set Value $R(Dmax)}
                if {$Value < $R(Dmin)} {set Value $R(Dmin)}

                set vl [rtos [dtoa $Value $R(M) $R(C)] $R(q) $R(u)]
                   if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME Read: $vl $Value $R(Amax) $R(Amin) $R(Dmax) $R(Dmin) $R(M) $R(C) $R(q) $R(u)"}
                set Value $vl
            }
            bVME  {
                if {$R(p) == "wo"} then {
                    return -code error -errorcode 0x20004 -errorinfo "Register is not readable"
                } else {
                    if {$MIDAS_VMEAccessServer_Verbose} \
                        {puts "VME request a=$R(a) o=$R(o) w=$R(w) d=$R(d) m=$R(am) l=$R(l)"}

                    set Addr [expr $R(a) + $R(o)]
                    set DataWidth $R(d)
                    set AddrMod $R(am)

                    if {$::VME::TestMode != 0} {
                      set rc [list -1 0]
                    } else {
                        set rc [rdblkVME -a $Addr -w $DataWidth -m $AddrMod -l $R(l)]
                    }

                    if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

#      for the CAEN interfaces rdblkVME will use CAENVME_FIFOBLTReadCycle
#      if the BERR option is enabled the block transfer will be terminated by a VME Bus Error
#      Hence rc = -1 is a good completion. The number of bytes transferred will be second datum returned

                    if {[first $rc] < -1} {return -code error -errorcode $rc -errorinfo "VME access error"}

                    set Value $rc
                }
            }
            vmeExec  {
                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME request $Address"}

                set rc [rdVME $Address]

                if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

                if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

                set Value [second $rc]
            }
            vmeConfiguredState {
                set Value $HasBeenConfigured
            }
            vmeRestoredState {
                set Value $HasBeenRestored
            }
            vmeStatus {
                set Value $Status
            }
            vmeDebug {
                set Value $Debug
            }
            vmeAddr {
                set Value "$Address"
            }
            vmeData {
                set Value $Data
            }
            default {
                return -code error -errorcode 0x10004 -errorinfo "Invalid request"
            }
        }
        return $Value
    }


    proc BusWrite {Address AddressModifier DataWidth Value} {
        global MIDAS_VMEAccessServer_Verbose

        if {$MIDAS_VMEAccessServer_Verbose} \
             {puts "called method VMEAccessServer::BusWrite with parameters $Address $AddressModifier $DataWidth $Value"}
             
#    applicable VME Address Modifiers  (single cycle accesses only)

#     A32UD = 0x09,             /* A32 nonprivileged data access */
#     A32UP = 0x0A,             /* A32 nonprivileged program access */
#     A32SD = 0x0D,             /* A32 supervisory data access */
#     A32SP = 0x0E,             /* A32 supervisory program access */
#     A16U = 0x29,              /* A16 nonprivileged access */
#     A16S = 0x2D,              /* A16 supervisory access */
#     A24UD = 0x39,             /* A24 nonprivileged data access */
#     A24UP = 0x3A,             /* A24 nonprivileged program access */
#     A24SD = 0x3D,             /* A24 supervisory data access */
#     A24SP = 0x3E,             /* A24 supervisory program access */

        switch $AddressModifier {
            A16S  {set AddrMod 0x2d}
            A16 -
            A16U  {set AddrMod 0x29}
            A24 -
            A24UD {set AddrMod 0x39}
            A24UP {set AddrMod 0x3a}
            A24SD {set AddrMod 0x3d}
            A24SP {set AddrMod 0x3e}
            A32 -
            A32UD {set AddrMod 0x09}
            A32UP {set AddrMod 0x0a}
            A32SD {set AddrMod 0x0d}
            A32SP {set AddrMod 0x0e}
            default {
                return -code error -errorcode 0x10004 -errorinfo "Address Modifier is invalid"
            }
        }

        switch $DataWidth {
            D8   {set DWidth 8}
            D16  {set DWidth 16}
            D32  {set DWidth 32}
            default {
                return -code error -errorcode 0x10004 -errorinfo "Data Width is invalid"
            }
        }

                    if {$::VME::TestMode != 0} {
                      global VMEData
                      set VMEData($Address) $Value
                      set rc [list 0 [set VMEData($Address)]]
                    } else {
                        set rc [wtVME -a $Address -w $DWidth -m $AddrMod -d $Value]
                    }

        if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

        if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

        return [rpcvar::rpcvar int 0]
    }

    proc BusRead {Address AddressModifier DataWidth} {
        global MIDAS_VMEAccessServer_Verbose

        if {$MIDAS_VMEAccessServer_Verbose} \
             {puts "called method VMEAccessServer::BusRead with parameters $Address $AddressModifier $DataWidth"}

        switch $AddressModifier {
            A16S  {set AddrMod 0x2d}
            A16 -
            A16U  {set AddrMod 0x29}
            A24 -
            A24UD {set AddrMod 0x39}
            A24UP {set AddrMod 0x3a}
            A24SD {set AddrMod 0x3d}
            A24SP {set AddrMod 0x3e}
            A32 -
            A32UD {set AddrMod 0x09}
            A32UP {set AddrMod 0x0a}
            A32SD {set AddrMod 0x0d}
            A32SP {set AddrMod 0x0e}
            default {
                return -code error -errorcode 0x10004 -errorinfo "Address Modifier is invalid"
            }
        }

        switch $DataWidth {
            D8   {set DWidth 8}
            D16  {set DWidth 16}
            D32  {set DWidth 32}
            default {
                return -code error -errorcode 0x10004 -errorinfo  "Data Width is invalid"
            }
        }

                   if {$::VME::TestMode != 0} {
                      global VMEData
                      if {![info exists VMEData($Address))]} {set VMEData($Address) 0x0003dead}
                      set rc [list 0 [set VMEData($Address)]]
                    } else {
                        set rc [rdVME -a $Address -w $DWidth -m $AddrMod]
                    }

        if {$MIDAS_VMEAccessServer_Verbose}  {puts "VME reply $rc"}

        if {[first $rc] < 0} {return -code error -errorcode $rc -errorinfo "VME access error"}

        set Value [second $rc];
        set Value [format "0x%0*x" [expr $DWidth/4] $Value]

        return [rpcvar::rpcvar int $Value]
    }


    proc InterfaceStatus {} {

        if {$::VME::TestMode == 0} {
           set rc [stVME]    ;#  get interface specific response 
        } else {
           set rc 0
        }

        return [rpcvar::rpcvar string "$rc"]
    }

    proc InterfaceReset {} {

        if {$::VME::TestMode == 0} {
           set rc [rstVME]
        } else {
           set rc 0
        }

        return [rpcvar::rpcvar string "$rc"]
    }


    # We have to publish the public methods...

     SOAP::export AccessServerVerbose
     SOAP::export DefineRegister
     SOAP::export InquireRegisters
     SOAP::export ReadRegisterAttributes
     SOAP::export ReadRegistersAttributes
     SOAP::export DeleteRegister
     SOAP::export DeleteRegisters
     SOAP::export WriteRegisterAttributes
     SOAP::export WriteRegistersAttributes
     SOAP::export WriteRegister
     SOAP::export WriteRegisters
     SOAP::export InitialiseRegister
     SOAP::export InitialiseRegisters
     SOAP::export ReadRegister
     SOAP::export ReadRegisters
     SOAP::export BusWrite
     SOAP::export BusRead
     SOAP::export InterfaceStatus
     SOAP::export InterfaceReset

     
}

# register this service with tclhttpd
puts [SOAP::Domain::register  -prefix  /VMEAccessServer  -namespace urn:VMEAccessServer -uri urn:VMEAccessServer]


proc ::SOAP::Domain::MIDAS_SOAPDomainHandler_/VMEAccessServer {optsname sock args} {   \
    ::SOAP::Domain::MIDAS_SOAPDomainHandler_/DefaultAccessServer $optsname $sock $args}



package provide VMEAccessServer 1.0
